# -*- coding: utf-8 -*-

"""
2021年还在由活跃的社区维护的rsa扩展有 ``cryptography`` 和 ``rsa``. cryp 集合了非常
多主流的加密算法, 并且都做了非常多的安全和性能优化(C语言实现). 而 rsa 是纯 Python
实现, 安装非常方便, 使用起来也较为简单. 所以本模块最后决定基于 rsa 进行开发.
"""

import typing as T

import rsa

from .cipher import BaseCipher

ERROR_MESSAGE = (
    "Asymmetric Encryption is for safely transfer secret key only! "
    "It is not suitable for file encryption!"
)


class AsymmetricCipher(BaseCipher):
    """
    A asymmetric encryption algorithm utility class helps you easily
    encrypt/decrypt text and files.

    :param my_pubkey: your public key
    :param my_privkey: your private key
    :param his_pubkey: other's public key you use to encrypt message

    **中文文档**

    非对称加密器. 主要用于加密少量信息. 通常用于安全地交换秘钥, 然后用该秘钥作为对称加密
    的钥匙对大量数据进行加密.
    """

    # key length/max length msg, 512/53, 1024/117, 2045/245
    _encrypt_chunk_size = 53
    _decrypt_chunk_size = 53

    def __init__(
        self,
        my_pubkey,
        my_privkey,
        his_pubkey,
    ):
        self.my_pubkey = my_pubkey
        self.my_privkey = my_privkey
        self.his_pubkey = his_pubkey
        self.sign = None
        self.password = None

    @staticmethod
    def new_keys(nbits=1024):
        """
        Create a new pair of public and private key pair to use.
        """
        public_key, private_key = rsa.newkeys(nbits, poolsize=1)
        return public_key, private_key

    newkeys = new_keys  # for backward compatibility

    def encrypt(
        self,
        binary: bytes,
        use_sign: bool = True,
        sign_method: str = "SHA-256",
        *args,
        **kwargs,
    ) -> bytes:
        """
        Encrypt binary data.

        :param sing_method: one of 'MD5', 'SHA-1', 'SHA-224', SHA-256',
            'SHA-384' or 'SHA-512'

        **中文文档**

        - 发送消息时只需要对方的 public_key
        - 如需使用签名, 则双方都需要持有对方的 public_key
        """
        token = rsa.encrypt(binary, self.his_pubkey)  # encrypt it
        if use_sign:
            self.sign = rsa.sign(binary, self.my_privkey, sign_method)  # sign it
        return token

    def decrypt(
        self,
        token: bytes,
        signature: T.Optional[bytes] = None,
        *args,
        **kwargs,
    ) -> bytes:
        """
        Decrypt binary data.

        **中文文档**

        - 接收消息时只需要自己的 private_key
        - 如需使用签名, 则双方都需要持有对方的 public_key
        """
        binary = rsa.decrypt(token, self.my_privkey)
        if signature:
            rsa.verify(binary, signature, self.his_pubkey)
        return binary

    def encrypt_file(self, **kwargs):  # pragma: no cover
        raise NotImplementedError(ERROR_MESSAGE)

    def decrypt_file(self, **kwargs):  # pragma: no cover
        raise NotImplementedError(ERROR_MESSAGE)

    def encrypt_dir(self, **kwargs):  # pragma: no cover
        raise NotImplementedError(ERROR_MESSAGE)

    def decrypt_dir(self, **kwargs):  # pragma: no cover
        raise NotImplementedError(ERROR_MESSAGE)



# -*- coding: utf-8 -*-

from windtalker.asymmetric import AsymmetricCipher


class TestAsymmetricCipher:
    def test_encrypt_decrypt_data(self):
        data = "Turn right at blue tree".encode("utf-8")
        A_pubkey, A_privkey = AsymmetricCipher.new_keys()
        B_pubkey, B_privkey = AsymmetricCipher.new_keys()

        cipherA = AsymmetricCipher(A_pubkey, A_privkey, B_pubkey)
        cipherB = AsymmetricCipher(B_pubkey, B_privkey, A_pubkey)

        token = cipherA.encrypt(data)
        sign = cipherA.sign

        data_new = cipherB.decrypt(token, signature=sign)
        assert data == data_new
        assert data != token


if __name__ == "__main__":
    from windtalker.tests import run_cov_test

    run_cov_test(__file__, "windtalker.asymmetric", preview=False)



# -*- coding: utf-8 -*-

import pytest
from pathlib_mate import Path
from windtalker import files


def test_get_encrypted_path_get_decrypted_path():
    p_before = Path(__file__)
    p_after = files.get_encrypted_path(p_before)
    assert p_after.abspath.endswith(files.DEFAULT_SUFFIX + ".py")
    assert files.get_decrypted_path(p_after) == p_before

    dir_before = p_before.parent
    dir_after = files.get_encrypted_path(dir_before)
    assert dir_after.abspath.endswith(files.DEFAULT_SUFFIX)
    assert files.get_decrypted_path(dir_after) == dir_before

    files.transform(
        p_before,
        p_after,
        converter=lambda x: x,
        overwrite=True,
        stream=True,
    )
    p_after.remove_if_exists()


def test_process_dst_overwrite_args():
    p_src = Path("test.txt")
    p_src, p_dst = files.process_dst_overwrite_args(
        p_src,
        dst=None,
        overwrite=False,
        src_to_dst_func=files.get_encrypted_path,
    )
    assert p_dst == files.get_encrypted_path(p_src)

    src = "test.txt"
    with pytest.raises(EnvironmentError):
        p_src, p_dst = files.process_dst_overwrite_args(
            src,
            dst=__file__,
            overwrite=False,
            src_to_dst_func=files.get_encrypted_path,
        )


if __name__ == "__main__":
    from windtalker.tests import run_cov_test

    run_cov_test(__file__, "windtalker.files", preview=False)



# -*- coding: utf-8 -*-

import typing as T
import os
import time
import base64

from pathlib_mate import Path, T_PATH_ARG

from . import files


class BaseCipher:
    """
    Base cipher class.

    Any cipher utility class that using any encryption algorithm can be
    inherited from this base class. The only method you need to implement is
    :meth:`BaseCipher.encrypt` and :meth:`BaseCipher.decrypt`. Because once you
    can encrypt binary data, then you can encrypt text, file, and directory.
    """

    _encrypt_chunk_size = 1024
    _decrypt_chunk_size = 1024

    def b64encode_str(self, text: str) -> str:
        """
        base64 encode a text, return string also.
        """
        return base64.b64encode(text.encode("utf-8")).decode("utf-8")

    def b64decode_str(self, text: str) -> str:
        """
        base64 decode a text, return string also.
        """
        return base64.b64decode(text.encode("utf-8")).decode("utf-8")

    def encrypt(self, binary: bytes, *args, **kwargs) -> bytes:
        """
        Overwrite this method using your encrypt algorithm.

        :param binary: binary data you need to encrypt

        :return: encrypted_binary, encrypted binary data
        """
        raise NotImplementedError

    def decrypt(self, binary: bytes, *args, **kwargs) -> bytes:
        """
        Overwrite this method using your decrypt algorithm.

        :param binary: binary data you need to decrypt

        :return: decrypted_binary, decrypted binary data
        """
        raise NotImplementedError

    def encrypt_binary(self, binary: bytes, *args, **kwargs) -> bytes:
        """ """
        return self.encrypt(binary, *args, **kwargs)

    def decrypt_binary(self, binary: bytes, *args, **kwargs) -> bytes:
        """ """
        return self.decrypt(binary, *args, **kwargs)

    def encrypt_text(self, text: str, *args, **kwargs) -> str:
        """
        Encrypt a string.
        """
        b = text.encode("utf-8")
        token = self.encrypt(b, *args, **kwargs)
        return base64.b64encode(token).decode("utf-8")

    def decrypt_text(self, text: str, *args, **kwargs) -> str:
        """
        Decrypt a string.
        """
        b = text.encode("utf-8")
        token = base64.b64decode(b)
        return self.decrypt(token, *args, **kwargs).decode("utf-8")

    def _show(
        self,
        message: str,
        indent: int = 0,
        enable_verbose: bool = True,
    ):  # pragma: no cover
        """Message printer."""
        if enable_verbose:
            print("    " * indent + message)

    def encrypt_file(
        self,
        path: T_PATH_ARG,
        output_path: T.Optional[T_PATH_ARG] = None,
        overwrite: bool = False,
        stream: bool = True,
        enable_verbose: bool = True,
        **kwargs,
    ):
        """
        Encrypt a file. If output_path are not given, then try to use the
        path with a surfix appended. The default automatical file path handling
        is defined here :meth:`windtalker.files.get_encrypted_file_path`

        :param path: path of the file you need to encrypt
        :param output_path: encrypted file output path
        :param overwrite: if True, then silently overwrite output file if exists
        :param stream: if it is a very big file, stream mode can avoid using
          too much memory
        :param enable_verbose: trigger on/off the help information
        """
        path, output_path = files.process_dst_overwrite_args(
            src=path,
            dst=output_path,
            overwrite=overwrite,
            src_to_dst_func=files.get_encrypted_path,
        )

        self._show("Encrypt '%s' ..." % path, enable_verbose=enable_verbose)
        st = time.process_time()
        files.transform(
            path,
            output_path,
            converter=self.encrypt,
            overwrite=overwrite,
            stream=stream,
            chunksize=self._encrypt_chunk_size,
        )
        self._show(
            "    Finished! Elapse %.6f seconds" % (time.process_time() - st,),
            enable_verbose=enable_verbose,
        )

        return output_path

    def decrypt_file(
        self,
        path: T_PATH_ARG,
        output_path: T_PATH_ARG = None,
        overwrite: bool = False,
        stream: bool = True,
        enable_verbose: bool = True,
        **kwargs,
    ):
        """
        Decrypt a file. If output_path are not given, then try to use the
        path with a surfix appended. The default automatical file path handling
        is defined here :meth:`windtalker.files.recover_path`

        :param path: path of the file you need to decrypt
        :param output_path: decrypted file output path
        :param overwrite: if True, then silently overwrite output file if exists
        :param stream: if it is a very big file, stream mode can avoid using
          too much memory
        :param enable_verbose: boolean, trigger on/off the help information
        """
        path, output_path = files.process_dst_overwrite_args(
            src=path,
            dst=output_path,
            overwrite=overwrite,
            src_to_dst_func=files.get_decrypted_path,
        )

        self._show("Decrypt '%s' ..." % path, enable_verbose=enable_verbose)
        st = time.process_time()
        files.transform(
            path,
            output_path,
            converter=self.decrypt,
            overwrite=overwrite,
            stream=stream,
            chunksize=self._decrypt_chunk_size,
        )
        self._show(
            "    Finished! Elapse %.6f seconds" % (time.process_time() - st,),
            enable_verbose=enable_verbose,
        )

        return output_path

    def encrypt_dir(
        self,
        path: T_PATH_ARG,
        output_path: T.Optional[T_PATH_ARG] = None,
        overwrite: bool = False,
        stream: bool = True,
        enable_verbose: bool = True,
    ):
        """
        Encrypt everything in a directory.

        :param path: path of the dir you need to encrypt
        :param output_path: encrypted dir output path
        :param overwrite: if True, then silently overwrite output file if exists
        :param stream: if it is a very big file, stream mode can avoid using
          too much memory
        :param enable_verbose: boolean, trigger on/off the help information
        """
        path, output_path = files.process_dst_overwrite_args(
            src=path,
            dst=output_path,
            overwrite=overwrite,
            src_to_dst_func=files.get_encrypted_path,
        )

        self._show(
            "--- Encrypt directory '%s' ---" % path, enable_verbose=enable_verbose
        )
        st = time.process_time()
        for current_dir, _, file_list in os.walk(path.abspath):
            new_dir = output_path.joinpath(Path(current_dir).relative_to(path))
            new_dir.mkdir_if_not_exists()
            for basename in file_list:
                old_path = os.path.join(current_dir, basename)
                new_path = new_dir.joinpath(basename)
                self.encrypt_file(
                    old_path,
                    new_path,
                    overwrite=overwrite,
                    stream=stream,
                    enable_verbose=enable_verbose,
                )
        self._show(
            "Complete! Elapse %.6f seconds" % (time.process_time() - st,),
            enable_verbose=enable_verbose,
        )
        return output_path

    def decrypt_dir(
        self,
        path: T_PATH_ARG,
        output_path: T.Optional[T_PATH_ARG] = None,
        overwrite: bool = False,
        stream: bool = True,
        enable_verbose: bool = True,
    ):
        """
        Decrypt everything in a directory.

        :param path: path of the dir you need to decrypt
        :param output_path: decrypted dir output path
        :param overwrite: if True, then silently overwrite output file if exists
        :param stream: if it is a very big file, stream mode can avoid using
          too much memory
        :param enable_verbose: boolean, trigger on/off the help information
        """
        path, output_path = files.process_dst_overwrite_args(
            src=path,
            dst=output_path,
            overwrite=overwrite,
            src_to_dst_func=files.get_decrypted_path,
        )

        self._show(
            "--- Decrypt directory '%s' ---" % path, enable_verbose=enable_verbose
        )
        st = time.process_time()
        for current_dir, _, file_list in os.walk(path):
            new_dir = output_path.joinpath(Path(current_dir).relative_to(path))
            new_dir.mkdir_if_not_exists()
            for basename in file_list:
                old_path = os.path.join(current_dir, basename)
                new_path = new_dir.joinpath(basename)
                self.decrypt_file(
                    old_path,
                    new_path,
                    overwrite=overwrite,
                    stream=stream,
                    enable_verbose=enable_verbose,
                )
        self._show(
            "Complete! Elapse %.6f seconds" % (time.process_time() - st,),
            enable_verbose=enable_verbose,
        )

        return output_path



# -*- coding: utf-8 -*-

"""
File handling utility functions.
"""

import typing as T
from pathlib_mate import Path, T_PATH_ARG

DEFAULT_SUFFIX = "-encrypted"  # windtalker secret file or folder suffix


def get_encrypted_path(
    original_path: T_PATH_ARG,
    suffix: str = DEFAULT_SUFFIX,
) -> Path:
    """
    Find the output encrypted file /dir path (by adding a suffix).

    Example:

    - file: ``${home}/test.txt`` -> ``${home}/test-encrypted.txt``
    - dir: ``${home}/Documents`` -> ``${home}/Documents-encrypted``
    """
    p = Path(original_path).absolute()
    return p.change(new_fname=p.fname + suffix)


def get_decrypted_path(
    encrypted_path: T_PATH_ARG,
    suffix: str = DEFAULT_SUFFIX,
) -> Path:
    """
    Find the original path of encrypted file or dir.

    Example:

    - file: ``${home}/test-encrypted.txt`` -> ``${home}/test.txt``
    - dir: ``${home}/Documents-encrypted`` -> ``${home}/Documents``
    """
    p = Path(encrypted_path).absolute()
    if not p.fname.endswith(suffix):
        raise ValueError(
            f"'{p}' is not a encrypted file or dir path "
            f"(filename not ends with suffix {suffix}."
        )
    return p.change(new_fname=p.fname[: -len(suffix)])


def transform(
    src: T_PATH_ARG,
    dst: T_PATH_ARG,
    converter: T.Callable,
    overwrite: bool = False,
    stream: bool = True,
    chunksize: int = 1024**2,
    **kwargs,
):
    """
    A file stream transform IO utility function.

    :param src: original file path
    :param dst: destination file path
    :param converter: binary content converter function
    :param overwrite: default False,
    :param stream: default True, if True, use stream IO mode, chunksize has to
      be specified.
    :param chunksize: default 1MB
    """
    p_src = Path(src).absolute()
    p_dst = Path(dst).absolute()
    if not overwrite:  # pragma: no cover
        if p_dst.exists():
            raise EnvironmentError(f"'{p_dst}' already exists!")

    with p_src.open("rb") as f_input:
        with p_dst.open("wb") as f_output:
            if stream:
                # fix chunksize to a reasonable range
                if chunksize > 1024**2 * 10:
                    chunksize = 1024**2 * 10
                elif chunksize < 1024**2:
                    chunksize = 1024**2

                # write file
                while 1:
                    content = f_input.read(chunksize)
                    if content:
                        f_output.write(converter(content, **kwargs))
                    else:
                        break
            else:  # pragma: no cover
                f_output.write(converter(f_input.read(), **kwargs))


def process_dst_overwrite_args(
    src: T_PATH_ARG,
    dst: T.Optional[T_PATH_ARG] = None,
    overwrite: bool = True,
    src_to_dst_func: T.Optional[T.Callable] = None,
) -> T.Tuple[Path, Path]:
    """
    Check when overwrite is not allowed, whether the destination exists.
    """
    p_src = Path(src).absolute()

    if dst is None:
        p_dst = src_to_dst_func(p_src)
    else:
        p_dst = Path(dst).absolute()

    if not overwrite:
        if p_dst.exists():
            raise EnvironmentError(f"output path '{p_dst}' already exists..")

    return p_src, p_dst



# -*- coding: utf-8 -*-

from windtalker import api


def test():
    _ = api
    _ = api.BaseCipher
    _ = api.AsymmetricCipher
    _ = api.SymmetricCipher


if __name__ == "__main__":
    from windtalker.tests import run_cov_test

    run_cov_test(__file__, "windtalker.api", preview=False)



# -*- coding: utf-8 -*-

import pytest

from windtalker.symmetric import SymmetricCipher
from windtalker.exc import PasswordError
from windtalker.tests import BaseTestCipher


class TestSymmetricCipher(BaseTestCipher):
    cipher = SymmetricCipher(password="MyPassword")
    with pytest.raises(ValueError):
        cipher.set_encrypt_chunk_size(0)
    cipher.set_encrypt_chunk_size(10 * 1024 * 1024)

    _ = cipher.metadata

    def test_decrypt_with_password(self):
        encrypted_text = self.cipher.encrypt_text("Hello World")
        cipher = SymmetricCipher(password="MyPassword")
        cipher.set_password(password="AnotherPassword")
        with pytest.raises(PasswordError):
            cipher.decrypt_text(encrypted_text)


if __name__ == "__main__":
    from windtalker.tests import run_cov_test

    run_cov_test(__file__, "windtalker.symmetric", preview=False)



# -*- coding: utf-8 -*-

import base64

from windtalker.cipher import BaseCipher
from windtalker.tests import BaseTestCipher


class MyCipher(BaseCipher):
    def encrypt(self, binary, *args, **kwargs):
        return (base64.b64encode(binary).decode("utf-8") + "X").encode("utf-8")

    def decrypt(self, binary, *args, **kwargs):
        return base64.b64decode(binary.decode("utf-8")[:-1].encode("utf-8"))


c = MyCipher()


def test_base64():
    s = "hello"
    assert c.b64decode_str(c.b64encode_str(s)) == s


class TestBaseCipher(BaseTestCipher):
    cipher = c


if __name__ == "__main__":
    from windtalker.tests import run_cov_test

    run_cov_test(__file__, "windtalker.cipher", preview=False)



# -*- coding: utf-8 -*-

import typing as T
import base64

from cryptography.fernet import Fernet
from .vendor.hashes import hashes

from .exc import PasswordError
from .paths import path_windtalker
from .cipher import BaseCipher


def read_windtalker_password():  # pragma: no cover
    return path_windtalker.read_text(encoding="utf-8").strip()


class SymmetricCipher(BaseCipher):
    """
    A symmetric encryption algorithm utility class helps you easily
    encrypt/decrypt text, files and even a directory.

    :param password: The secret password you use to encrypt all your message.
      If you feel uncomfortable to put that in your code, you can leave it
      empty. The system will ask you manually enter that later.

    **中文文档**

    对称加密器。
    """

    _encrypt_chunk_size = 1024 * 1024  # 1 MB
    _decrypt_chunk_size = 1398200  # 1.398 MB
    """Symmtric algorithm needs to break big files in small chunk, and encrypt
    them one by one, and concatenate them at the end. Each chunk has a fixed 
    size. That's what these two attributes for.
    """

    def __init__(self, password: T.Optional[str] = None):
        if password:
            fernet_key = self.any_text_to_fernet_key(password)
            self.fernet = Fernet(fernet_key)  # type: Fernet
        else:  # pragma: no cover
            if path_windtalker.exists():
                self.set_password(read_windtalker_password())
            else:
                self.input_password()

    def any_text_to_fernet_key(self, text: str) -> bytes:
        """
        Convert any text to a fernet key for encryption.
        """
        md5 = hashes.of_str(text)
        fernet_key = base64.b64encode(md5.encode("utf-8"))
        return fernet_key

    def input_password(self):  # pragma: no cover
        """
        Manually enter a password for encryption on keyboard.
        """
        password = input("Please enter your secret key (case sensitive): ")
        self.set_password(password)

    def set_password(self, password: str):
        """
        Set a new password for encryption.
        """
        self.__init__(password=password)

    def set_encrypt_chunk_size(self, size: int):
        if 1024 * 1024 < size < 100 * 1024 * 1024:
            self._encrypt_chunk_size = size
            self._decrypt_chunk_size = len(self.encrypt(b"x" * size))
        else:
            raise ValueError(
                f"Cannot set encrypt chunk size = {size}, "
                f"encrypt chunk size has to be between 1MB and 100MB"
            )

    @property
    def metadata(self) -> dict:
        return {
            "_encrypt_chunk_size": self._encrypt_chunk_size,
            "_decrypt_chunk_size": self._decrypt_chunk_size,
        }

    def encrypt(self, binary: bytes, *args, **kwargs) -> bytes:
        """
        Encrypt binary data.
        """
        return self.fernet.encrypt(binary)

    def decrypt(self, binary: bytes, *args, **kwargs) -> bytes:
        """
        Decrypt binary data.
        """
        try:
            return self.fernet.decrypt(binary)
        except:
            raise PasswordError("Ops, wrong magic word!")
